# -*- coding: utf-8 -*-
"""
Classes to handle multiple years of resource data

Data split by time in chunks of less than a year can be opened by the
MultiTimeResource class, but not by these classes.
"""
import pandas as pd
import numpy as np
import os

from rex.multi_file_resource import MULTI_FILE_CLASS_MAP, MultiFileResource
from rex.multi_time_resource import MultiTimeH5, MultiTimeResource
from rex.renewable_resource import (NSRDB, SolarResource, WindResource,
                                    WaveResource)
from rex.resource import Resource
from rex.utilities.parse_keys import parse_slice
from rex.utilities.utilities import parse_year


class MultiYearH5(MultiTimeH5):
    """
    Class to handle multiple years of h5 Resources
    """

    def __init__(self, h5_path, years=None, res_cls=Resource, hsds=False,
                 hsds_kwargs=None, **res_cls_kwargs):
        """
        Parameters
        ----------
        h5_path : str
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be a path
            on HSDS starting with /nrel/ or a path on s3 starting with s3://
        years : list, optional
            List of integer years to access, by default None
        res_cls : obj
            Resource class to use to open and access resource data
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False. This is now redundant; file paths
            starting with /nrel/ will be treated as hsds=True by default
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        self.h5_path = h5_path
        self._file_paths = self._get_file_paths(h5_path, hsds=hsds,
                                                hsds_kwargs=hsds_kwargs)
        self._file_paths, self._years = self._get_years(self._file_paths,
                                                        years)
        res_cls_kwargs.update({'hsds': hsds})
        self._h5_map = self._map_file_instances(self._file_paths,
                                                self._years,
                                                res_cls=res_cls,
                                                **res_cls_kwargs)
        self._years = self._h5_map['year'].values.tolist()

        self._datasets = None
        self._shape = None
        self._time_index = None

    def __repr__(self):
        msg = ("{} for {}:\n Contains data for {} years"
               .format(self.__class__.__name__, self.h5_path, len(self)))
        return msg

    def __len__(self):
        return len(set(self.years))

    def __getitem__(self, year):
        if isinstance(year, str):
            year = int(year)

        if year not in self._h5_map['year'].values:
            raise ValueError('{} is invalid, must be one of: {}'
                             .format(year, self.years))

        idx = np.where(self._h5_map['year'] == year)[0][0]
        h5 = self._h5_map.at[idx, 'h5']

        return h5

    def __iter__(self):
        return iter(self.years)

    def __contains__(self, year):
        return year in self.years

    @property
    def years(self):
        """
        Available years ordered the same way as self.files

        Returns
        -------
        list
            List of dataset present in .h5 files
        """
        return self._years

    @property
    def files(self):
        """
        Available file paths ordered the same way as self.years

        Returns
        -------
        list
        """
        return self._file_paths

    @property
    def time_index(self):
        """
        Multi-year datetime index

        Returns
        -------
        pandas.DatatimeIndex
        """
        if self._time_index is None:
            for h5 in self._h5_map['h5'].unique():
                if self._time_index is None:
                    self._time_index = h5.time_index
                else:
                    ti = self._time_index.append(h5.time_index)
                    self._time_index = ti

        return self._time_index

    @staticmethod
    def _map_file_instances(file_paths, years, res_cls=Resource,
                            **res_cls_kwargs):
        """
        Open all .h5 files and map the open h5py instances to the
        associated file paths

        Parameters
        ----------
        file_paths : list
            List of filepaths for this handler to handle.

        Returns
        -------
        h5_map : pd.DataFrame
            DataFrame mapping file paths to open resource instances and
            datasets per file (columns: fp, h5, and dsets)
        """

        h5_map = pd.DataFrame({'fp': file_paths, 'year': years, 'h5': None})
        h5_map = h5_map.sort_values('year').reset_index(drop=True)

        if len(h5_map['year'].unique()) < len(h5_map):
            del res_cls_kwargs['hsds']  # no multi file res on hsds
            for _, subdf in h5_map.groupby('year'):
                fps = subdf['fp'].values.tolist()
                handle = MULTI_FILE_CLASS_MAP.get(res_cls, MultiFileResource)
                h5 = handle(fps, **res_cls_kwargs)
                for i in subdf.index:
                    h5_map.at[i, 'h5'] = h5

        else:
            for i, f_path in enumerate(h5_map['fp']):
                h5_map.at[i, 'h5'] = res_cls(f_path, **res_cls_kwargs)

        h5_map['dsets'] = [h5.dsets for h5 in h5_map['h5'].values]

        return h5_map

    @staticmethod
    def _get_years(file_paths, years):
        """Reduce file path list to requested years and/or return list of
        years corresponding to file paths

        Parameters
        ----------
        file_paths : list
            List of filepaths for this handler to handle.
        years : list | None
            List of years of interest. Should be a subset of years in file_map.
            Can also be None for all years found by the h5_path input.

        Returns
        -------
        file_paths : list
            List of filepaths for this handler to handle.
        years : list
            List of integer years corresponding to the file_paths list
        """

        fp_years = [int(parse_year(os.path.basename(fp), option='raise'))
                    for fp in file_paths]

        if years is None:
            years = fp_years
        elif any(int(y) not in fp_years for y in years):
            years = sorted([int(y) for y in years])
            raise RuntimeError('Requested years "{}" not all found in '
                               'file years "{}"'.format(years, fp_years))
        else:
            filtered_fps = []
            years = sorted([int(y) for y in years])
            for target_year in years:
                for fp_year, fp in zip(fp_years, file_paths):
                    if int(target_year) == int(fp_year):
                        filtered_fps.append(fp)
                        break
            file_paths = filtered_fps

        return file_paths, years

    @staticmethod
    def _check_for_years(time_slice):
        """
        Check to see if temporal slice is a year (str) or list of years (strs)
        to extract data for

        Parameters
        ----------
        time_slice : list | slice | int | str
            Temporal slice to extract

        Returns
        -------
        check : bool
            True if temporal slice is a year (str) or list of years (strs),
            else False
        """
        check = False
        if isinstance(time_slice, (list, tuple)):
            time_slice = time_slice[0]

        if isinstance(time_slice, str):
            check = True

        return check

    def year_index(self, year):
        """
        Extract time_index for a specific year

        Parameters
        ----------
        year : int
            Year to extract time_index for

        Returns
        -------
        time_index : pandas.DatetimeIndex
            Resource datetime index for desired year
        """
        return self.time_index[self.time_index.year == year]

    def _get_ds(self, ds_name, ds_slice):
        """
        Extract data from given dataset

        Parameters
        ----------
        ds_name : str
            Variable dataset to be extracted
        ds_slice : int | list | slice
            tuple describing slice of dataset array to extract

        Returns
        -------
        out : ndarray
            ndarray of variable timeseries data
            If unscale, returned in native units else in scaled units
        """
        ds_slice = parse_slice(ds_slice)
        out = []
        time_slice = ds_slice[0]
        if self._check_for_years(time_slice):
            years = time_slice
            year_slice = (slice(None), ) + ds_slice[1:]
            if isinstance(years, str):
                years = [years]

            for year in years:
                year = int(year)
                out.append(self[year]._get_ds(ds_name, year_slice))

            out = np.concatenate(out, axis=0)

        elif isinstance(time_slice, (int, np.integer)):
            time_step = self.time_index[time_slice]
            year = time_step.year
            year_index = self.year_index(year)
            year_slice = np.where(time_step == year_index)[0][0]
            year_slice = (year_slice, ) + ds_slice[1:]
            out = self[year]._get_ds(ds_name, year_slice)

        else:
            time_index = self.time_index[time_slice]
            year_map = time_index.year
            for year in year_map.unique():
                year_index = self.year_index(year)
                year_slice = year_index.isin(time_index[year_map == year])
                year_slice = \
                    self._check_time_slice(np.where(year_slice)[0])
                year_slice = (year_slice, ) + ds_slice[1:]
                out.append(self[year]._get_ds(ds_name, year_slice))

            out = np.concatenate(out, axis=0)

        return out

    def close(self):
        """
        Close all h5py.File instances
        """
        for f in self._h5_map['h5']:
            f.close()


class MultiYearResource(MultiTimeResource):
    """
    Class to handle multiple years of resource data stored accross multiple
    .h5 files. This also works if each year is split into multiple files each
    containing different datasets (e.g. for Sup3rCC and hi-res WTK+NSRDB). Data
    split by time in chunks of less than a year can be opened by the
    MultiTimeResource class, but not by this class.

    Note that files across years must have the same meta data, and files within
    the same year must have the same meta and time_index.

    Examples
    --------

    Extracting the resource's Datetime Index

    >>> path = '$TESTDATADIR/nsrdb/ri_100_nsrdb_*.h5'
    >>> with MultiYearResource(path) as res:
    >>>     ti = res.time_index
    >>>
    >>> ti
    DatetimeIndex(['2012-01-01 00:00:00', '2012-01-01 00:30:00',
                   '2012-01-01 01:00:00', '2012-01-01 01:30:00',
                   '2012-01-01 02:00:00', '2012-01-01 02:30:00',
                   '2012-01-01 03:00:00', '2012-01-01 03:30:00',
                   '2012-01-01 04:00:00', '2012-01-01 04:30:00',
                   ...
                   '2013-12-31 19:00:00', '2013-12-31 19:30:00',
                   '2013-12-31 20:00:00', '2013-12-31 20:30:00',
                   '2013-12-31 21:00:00', '2013-12-31 21:30:00',
                   '2013-12-31 22:00:00', '2013-12-31 22:30:00',
                   '2013-12-31 23:00:00', '2013-12-31 23:30:00'],
                  dtype='datetime64[ns]', length=35088, freq=None)

    NOTE: time_index covers data from 2012 and 2013

    >>> with MultiYearResource(path) as res:
    >>>     print(res.h5_files)

    ['/Users/mrossol/Git_Repos/rex/tests/data/nsrdb/ri_100_nsrdb_2012.h5',
     '/Users/mrossol/Git_Repos/rex/tests/data/nsrdb/ri_100_nsrdb_2013.h5']

    Data slicing works the same as with "Resource" except axis 0 now covers
    2012 and 2013

    >>> with MultiYearResource(path) as res:
    >>>     temperature = res['air_temperature']
    >>>
    >>> temperature
    [[ 4.  5.  5. ...  4.  3.  4.]
     [ 4.  4.  5. ...  4.  3.  4.]
     [ 4.  4.  5. ...  4.  3.  4.]
     ...
     [-1. -1.  0. ... -2. -3. -2.]
     [-1. -1.  0. ... -2. -3. -2.]
     [-1. -1.  0. ... -2. -3. -2.]]
    >>> temperature.shape
    (35088, 100)

    >>> with MultiYearResource(path) as res:
    >>>     temperature = res['air_temperature', ::100] # every 100th timestep
    >>>
    >>> temperature
    [[ 4.  5.  5. ...  4.  3.  4.]
     [ 1.  1.  2. ...  0.  0.  1.]
     [-2. -1. -1. ... -2. -4. -2.]
     ...
     [-3. -2. -2. ... -3. -4. -3.]
     [ 0.  0.  1. ...  0. -1.  0.]
     [ 3.  3.  3. ...  2.  2.  3.]]
    >>> temperature.shape
    (351, 100)

    You can also request a specific year of data using a string representation
    of the year of interest
    NOTE: you can also request a list of years using strings

    >>> with MultiYearResource(path) as res:
    >>>     temperature = res['air_temperature', '2012'] # every 100th timestep
    >>>
    >>> temperature
    [[4. 5. 5. ... 4. 3. 4.]
     [4. 4. 5. ... 4. 3. 4.]
     [4. 4. 5. ... 4. 3. 4.]
     ...
     [1. 1. 2. ... 0. 0. 0.]
     [1. 1. 2. ... 0. 0. 1.]
     [1. 1. 2. ... 0. 0. 1.]]
    >>> temperature.shape
    (17520, 100)
    """

    def __init__(self, h5_path, years=None, unscale=True, str_decode=True,
                 res_cls=Resource, hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards. Can also be a path on HSDS starting with
            /nrel/ or a path on s3 starting with s3://
        years : list, optional
            List of years to access, by default None
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        res_cls : obj
            Resource handler to us to open individual .h5 files
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False. This is now redundant; file paths
            starting with /nrel/ will be treated as hsds=True by default
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        self.h5_path = h5_path
        self._time_index = None
        # Map variables to their .h5 files
        cls_kwargs = {'unscale': unscale, 'str_decode': str_decode,
                      'hsds': hsds, 'hsds_kwargs': hsds_kwargs}
        self._h5 = MultiYearH5(self.h5_path, years=years, res_cls=res_cls,
                               **cls_kwargs)
        self.h5_files = self._h5.h5_files
        self.h5_file = self.h5_files[0]

    @property
    def years(self):
        """
        Available years

        Returns
        -------
        list
            List of dataset present in .h5 files
        """
        return self.h5.years


class MultiYearSolarResource:
    """
    Class to handle multiple years of solar resource data stored accross
    multiple .h5 files
    """
    def __init__(self, h5_path, years=None, unscale=True, str_decode=True,
                 hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be a path
            on HSDS starting with /nrel/ or a path on s3 starting with s3://
        years : list, optional
            List of years to access, by default None
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False. This is now redundant; file paths
            starting with /nrel/ will be treated as hsds=True by default
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, years=years, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=SolarResource)


class MultiYearNSRDB(MultiYearResource):
    """
    Class to handle multiple years of NSRDB data stored accross
    multiple .h5 files
    """

    def __init__(self, h5_path, years=None, unscale=True, str_decode=True,
                 hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards. Can also be a path on HSDS starting with
            /nrel/ or a path on s3 starting with s3://
        years : list, optional
            List of years to access, by default None
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False. This is now redundant; file paths
            starting with /nrel/ will be treated as hsds=True by default
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, years=years, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=NSRDB)


class MultiYearWindResource(MultiYearResource):
    """
    Class to handle multiple years of wind resource data stored accross
    multiple .h5 files
    """

    def __init__(self, h5_path, years=None, unscale=True, str_decode=True,
                 hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards. Can also be a path on HSDS starting with
            /nrel/ or a path on s3 starting with s3://
        years : list, optional
            List of years to access, by default None
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False. This is now redundant; file paths
            starting with /nrel/ will be treated as hsds=True by default
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, years=years, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=WindResource)


class MultiYearWaveResource(MultiYearResource):
    """
    Class to handle multiple years of wave resource data stored accross
    multiple .h5 files
    """

    def __init__(self, h5_path, years=None, unscale=True, str_decode=True,
                 hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards. Can also be a path on HSDS starting with
            /nrel/ or a path on s3 starting with s3://
        years : list, optional
            List of years to access, by default None
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False. This is now redundant; file paths
            starting with /nrel/ will be treated as hsds=True by default
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, years=years, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=WaveResource)



# -*- coding: utf-8 -*-
"""
Classes to handle resource data stored over multiple files
"""
import os
from glob import glob
from itertools import chain
from fnmatch import fnmatch
import logging

import numpy as np
import pandas as pd

from rex.renewable_resource import (
    NSRDB,
    SolarResource,
    WaveResource,
    WindResource,
)
from rex.resource import Resource, BaseDatasetIterable
from rex.utilities.exceptions import FileInputError
from rex.utilities.parse_keys import parse_keys, parse_slice


logger = logging.getLogger(__name__)


class MultiTimeH5:
    """
    Class to handle h5 Resources stored over multiple temporal files
    """

    def __init__(self, h5_path, res_cls=Resource, hsds=False, hsds_kwargs=None,
                 **res_cls_kwargs):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards.
        res_cls : obj
            Resource class to use to open and access resource data
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        res_cls_kwargs : dict, optional
            Kwargs for `res_cls`
        """
        self.h5_path = h5_path
        self._file_paths = self._get_file_paths(h5_path, hsds=hsds,
                                                hsds_kwargs=hsds_kwargs)
        res_cls_kwargs.update({'hsds': hsds})
        self._h5_map = self._map_file_instances(self._file_paths,
                                                res_cls=res_cls,
                                                **res_cls_kwargs)

        self._datasets = None
        self._shape = None
        self._time_index = None
        self._time_slice_map = []

    def __repr__(self):
        msg = ("{} for {}:\n Contains data from {} files"
               .format(self.__class__.__name__, self.h5_path, len(self)))
        return msg

    def __getitem__(self, file):
        fn_fp_map = {os.path.basename(fp): fp for fp in self._file_paths}
        if file in self._h5_map['fp'].values:
            iloc = np.where(self._h5_map['fp'] == file)[0][0]
            h5 = self._h5_map.at[iloc, 'h5']
        elif file in fn_fp_map:
            iloc = np.where(self._h5_map['fp'] == fn_fp_map[file])[0][0]
            h5 = self._h5_map.at[iloc, 'h5']
        else:
            raise ValueError('{} is invalid, must be one of: {}'
                             .format(file, self._file_paths))

        return h5

    @property
    def attrs(self):
        """
        Global .h5 file attributes sourced from first .h5 file

        Returns
        -------
        attrs : dict
        """
        attrs = dict(self.h5.attrs)

        return attrs

    @property
    def files(self):
        """
        Available file paths

        Returns
        -------
        list
        """
        return sorted(self._file_paths)

    @property
    def h5_files(self):
        """
        .h5 files data is being sourced from

        Returns
        -------
        list
        """
        return sorted(self._h5_map['fp'])

    @property
    def h5(self):
        """
        open h5 file handler for a single .h5 file

        Returns
        -------
        h5py.File
        """
        return self._h5_map['h5'].values[0]

    @property
    def datasets(self):
        """
        Available datasets

        Returns
        -------
        list
        """
        if self._datasets is None:
            all_dsets = self._h5_map['dsets'].values.tolist()
            dsets = [d for sub in all_dsets for d in sub]
            self._datasets = list(set(dsets))

        return self._datasets

    @property
    def resource_datasets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        res_dsets = [ds for ds in self.datasets
                     if ds not in ['meta', 'time_index', 'coordinates']]

        return res_dsets

    @property
    def shape(self):
        """
        Dataset shape (time, sites)

        Returns
        -------
        tuple
        """
        if self._shape is None:
            self._shape = (len(self.time_index), self.h5.shape[1])

        return self._shape

    @property
    def time_index(self):
        """
        Multi-year datetime index

        Returns
        -------
        pandas.DatatimeIndex
        """
        if self._time_index is None:
            time_slice_map = []
            for _, row in self._h5_map.iterrows():
                h5 = row['h5']
                fp = row['fp']
                ti = h5.time_index
                time_slice_map.append(np.full(len(ti), os.path.basename(fp)))
                if self._time_index is None:
                    self._time_index = ti
                else:
                    self._time_index = self._time_index.append(ti)

            if len(self._time_index) != len(np.unique(self._time_index)):
                unique, duplicates = np.unique(self._time_index,
                                               return_counts=True)
                duplicates = np.where(duplicates > 1)[0]
                duplicates = unique[duplicates]
                msg = ('The combined time_index has {} duplicate values:\n{}'
                       .format(len(duplicates), duplicates))
                raise RuntimeError(msg)

            self._time_slice_map = np.concatenate(time_slice_map, axis=0)

        return self._time_index

    @staticmethod
    def _get_hsds_file_paths(h5_path, hsds_kwargs=None):
        """
        Get a list of h5 filepaths matching the h5_path specification from HSDS

        Parameters
        ----------
        h5_path : str
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes.
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None

        Returns
        -------
        file_paths : list
            List of filepaths for this handler to handle.
        """
        try:
            import h5pyd
        except Exception as e:
            msg = (f'Tried to open hsds file path: "{h5_path}" with '
                   'h5pyd but could not import, try '
                   '`pip install NREL-rex[hsds]`')
            logger.error(msg)
            raise ImportError(msg) from e

        if hsds_kwargs is None:
            hsds_kwargs = {}

        if isinstance(h5_path, (list, tuple)):
            msg = ('HSDS filepath must be a string, possibly with glob '
                   f'pattern, but received list/tuple: {h5_path}')
            logger.error(msg)
            raise TypeError(msg)

        hsds_dir = os.path.dirname(h5_path)
        fn = os.path.basename(h5_path)

        if '*' in hsds_dir:
            msg = ('HSDS path specifications cannot handle wildcards in the '
                   'directory name! The directory must be explicit but the '
                   'filename can have wildcards. This HSDS h5_path input '
                   'cannot be used: {}'.format(h5_path))
            logger.error(msg)
            raise FileNotFoundError(msg)

        if not fn:
            msg = ('h5_path must be a unix shell style pattern with '
                   'wildcard * in order to find files, but received '
                   'directory specification: {}'.format(h5_path))
            logger.error(msg)
            raise FileInputError(msg)

        with h5pyd.Folder(hsds_dir + '/', **hsds_kwargs) as f:
            file_paths = [f'{hsds_dir}/{fn}' for fn in f
                          if fnmatch(f'{hsds_dir}/{fn}', h5_path)]

        return file_paths

    @staticmethod
    def _get_s3_file_paths(h5_path):
        """
        Get a list of h5 filepaths matching the h5_path specification from s3

        Parameters
        ----------
        h5_path : str
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes.

        Returns
        -------
        file_paths : list
            List of filepaths for this handler to handle.
        """
        try:
            import s3fs
        except Exception as e:
            msg = (f'Tried to open s3 file path: "{h5_path}" with '
                   'fsspec but could not import, try '
                   '`pip install NREL-rex[s3]`')
            logger.error(msg)
            raise ImportError(msg) from e

        s3 = s3fs.S3FileSystem(anon=True)

        if isinstance(h5_path, (list, tuple)):
            file_paths = [s3.glob(fp) for fp in h5_path]
            file_paths = list(chain.from_iterable(file_paths))
        elif isinstance(h5_path, str):
            file_paths = s3.glob(h5_path)

        # s3fs glob drops prefix for some reason
        for i, fp in enumerate(file_paths):
            if not fp.startswith('s3://'):
                file_paths[i] = f's3://{fp}'

        return file_paths

    @classmethod
    def _get_file_paths(cls, h5_path, hsds=False, hsds_kwargs=None):
        """
        Get a file list based on the h5_path specification.

        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None

        Returns
        -------
        file_paths : list
            List of filepaths for this handler to handle.
        """

        if Resource.is_hsds_file(h5_path) or hsds:
            file_paths = cls._get_hsds_file_paths(h5_path,
                                                  hsds_kwargs=hsds_kwargs)
        elif Resource.is_s3_file(h5_path):
            file_paths = cls._get_s3_file_paths(h5_path)
        elif isinstance(h5_path, (list, tuple)):
            file_paths = list(chain.from_iterable(glob(fp) for fp in h5_path))
            for fp in file_paths:
                assert os.path.exists(fp), 'Does not exist: {}'.format(fp)
        elif os.path.isdir(h5_path):
            msg = ('h5_path must be a unix shell style pattern with '
                   'wildcard * in order to find files, but received '
                   'directory specification: {}'.format(h5_path))
            raise FileInputError(msg)
        elif isinstance(h5_path, str):
            file_paths = glob(h5_path)

        if not any(file_paths):
            msg = ('Could not find any file paths with pattern: {}'
                   .format(h5_path))
            raise FileInputError(msg)

        return file_paths

    @staticmethod
    def _map_file_instances(file_paths, res_cls=Resource, **res_cls_kwargs):
        """
        Open all .h5 files and map the open h5py instances to the
        associated file paths

        Parameters
        ----------
        file_paths : list
            List of filepaths for this handler to handle.

        Returns
        -------
        h5_map : pd.DataFrame
            DataFrame mapping file paths to open resource instances and
            datasets per file (columns: fp, h5, and dsets)
        """
        h5_map = pd.DataFrame({'fp': file_paths, 'h5': None,
                               't0': None, 't1': None})
        for i, f_path in enumerate(file_paths):
            h5_map.at[i, 'h5'] = res_cls(f_path, **res_cls_kwargs)
            h5_map.at[i, 't0'] = h5_map.at[i, 'h5'].time_index.values[0]
            h5_map.at[i, 't1'] = h5_map.at[i, 'h5'].time_index.values[1]

        h5_map['dsets'] = [h5.dsets for h5 in h5_map['h5'].values]
        h5_map = h5_map.sort_values('t0').reset_index(drop=True)

        return h5_map

    @staticmethod
    def _check_time_slice(time_slice):
        """
        Check to see if time positions can be represented as a slice

        Parameters
        ----------
        time_slice : ndarray | list
            List of temporal positions

        Returns
        -------
        time_slice : ndarray | list | slice
            Slice covering range of positions to extract if possible
        """
        s = time_slice[0]
        e = time_slice[-1] + 1
        if (e - s) == len(time_slice):
            time_slice = slice(s, e, None)

        return time_slice

    def _map_time_slice(self, time_slice):
        """
        Map timeslices to files

        Parameters
        ----------
        time_slice : int | list | slice
            tuple describing slice of dataset array to extract

        Returns
        -------
        file_times : dict
            Dictionary mapping files to the time_slices to extract
        """
        time_index = self.time_index[time_slice]
        files = self._time_slice_map[time_slice]
        file_times = {}
        for file in np.unique(files):
            ti = self[file].time_index
            file_slice = np.where(ti.isin(time_index))[0]
            file_slice = self._check_time_slice(file_slice)
            file_times[file] = file_slice

        return file_times

    def _get_ds(self, ds_name, ds_slice):
        """
        Extract data from given dataset

        Parameters
        ----------
        ds_name : str
            Variable dataset to be extracted
        ds_slice : int | list | slice
            tuple describing slice of dataset array to extract

        Returns
        -------
        out : ndarray
            ndarray of variable timeseries data
            If unscale, returned in native units else in scaled units
        """
        ds_slice = parse_slice(ds_slice)
        out = []
        time_slice = ds_slice[0]
        if isinstance(time_slice, (int, np.integer)):
            time_step = self.time_index[time_slice]
            file = self._time_slice_map[time_slice]
            time_index = self[file].time_index
            time_slice = np.where(time_step == time_index)[0][0]
            file_slice = (time_slice, ) + ds_slice[1:]
            out = self[file]._get_ds(ds_name, file_slice)
        else:
            file_times = self._map_time_slice(ds_slice[0])
            for file, time_slice in file_times.items():
                file_slice = (time_slice, ) + ds_slice[1:]
                out.append(self[file]._get_ds(ds_name, file_slice))

            out = np.concatenate(out, axis=0)

        return out

    def close(self):
        """
        Close all h5py.File instances
        """
        for f in self._h5_map['h5']:
            f.close()


class MultiTimeResource(BaseDatasetIterable):
    """
    Class to handle resource data stored temporally accross multiple
    .h5 files

    Examples
    --------
    Extracting the resource's Datetime Index

    >>> path = '$TESTDATADIR/nsrdb/ri_100_nsrdb_*.h5'
    >>> with MultiTimeResource(path) as res:
    >>>     ti = res.time_index
    >>>
    >>> ti
    DatetimeIndex(['2012-01-01 00:00:00', '2012-01-01 00:30:00',
                   '2012-01-01 01:00:00', '2012-01-01 01:30:00',
                   '2012-01-01 02:00:00', '2012-01-01 02:30:00',
                   '2012-01-01 03:00:00', '2012-01-01 03:30:00',
                   '2012-01-01 04:00:00', '2012-01-01 04:30:00',
                   ...
                   '2013-12-31 19:00:00', '2013-12-31 19:30:00',
                   '2013-12-31 20:00:00', '2013-12-31 20:30:00',
                   '2013-12-31 21:00:00', '2013-12-31 21:30:00',
                   '2013-12-31 22:00:00', '2013-12-31 22:30:00',
                   '2013-12-31 23:00:00', '2013-12-31 23:30:00'],
                  dtype='datetime64[ns]', length=35088, freq=None)

    NOTE: time_index covers data from 2012 and 2013

    >>> with MultiTimeResource(path) as res:
    >>>     print(res.h5_files)

    ['/Users/mrossol/Git_Repos/rex/tests/data/nsrdb/ri_100_nsrdb_2012.h5',
     '/Users/mrossol/Git_Repos/rex/tests/data/nsrdb/ri_100_nsrdb_2013.h5']

    Data slicing works the same as with "Resource" except axis 0 now covers
    2012 and 2013

    >>> with MultiTimeResource(path) as res:
    >>>     temperature = res['air_temperature']
    >>>
    >>> temperature
    [[ 4.  5.  5. ...  4.  3.  4.]
     [ 4.  4.  5. ...  4.  3.  4.]
     [ 4.  4.  5. ...  4.  3.  4.]
     ...
     [-1. -1.  0. ... -2. -3. -2.]
     [-1. -1.  0. ... -2. -3. -2.]
     [-1. -1.  0. ... -2. -3. -2.]]
    >>> temperature.shape
    (35088, 100)

    >>> with MultiTimeResource(path) as res:
    >>>     temperature = res['air_temperature', ::100] # every 100th timestep
    >>>
    >>> temperature
    [[ 4.  5.  5. ...  4.  3.  4.]
     [ 1.  1.  2. ...  0.  0.  1.]
     [-2. -1. -1. ... -2. -4. -2.]
     ...
     [-3. -2. -2. ... -3. -4. -3.]
     [ 0.  0.  1. ...  0. -1.  0.]
     [ 3.  3.  3. ...  2.  2.  3.]]
    >>> temperature.shape
    (351, 100)
    """

    def __init__(self, h5_path, unscale=True, str_decode=True,
                 res_cls=Resource, hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        res_cls : obj
            Resource handler to us to open individual .h5 files
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """

        self.h5_path = h5_path
        self._time_index = None
        # Map variables to their .h5 files
        cls_kwargs = {'unscale': unscale, 'str_decode': str_decode,
                      'hsds': hsds, 'hsds_kwargs': hsds_kwargs}
        self._h5 = MultiTimeH5(self.h5_path, res_cls=res_cls, **cls_kwargs)
        self.h5_files = self._h5.h5_files
        self.h5_file = self.h5_files[0]

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self.h5_path)
        return msg

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

        if type is not None:
            raise

    def __len__(self):
        return len(self.h5.time_index)

    def __getitem__(self, keys):
        ds, ds_slice = parse_keys(keys)

        if ds.endswith('time_index'):
            out = self.h5.h5._get_time_index(ds_slice)
        elif ds.endswith('meta'):
            out = self.h5.h5._get_meta(ds, ds_slice)
        elif ds.endswith('coordinates'):
            out = self.h5.h5._get_coords(ds, ds_slice)
        else:
            out = self.h5._get_ds(ds, ds_slice)

        return out

    def __contains__(self, dset):
        return dset in self.datasets

    @property
    def h5(self):
        """
        Open class instance that handles all .h5 files that data is to
        be extracted from

        Returns
        -------
        h5 : MultiTimeH5 | MultiYearH5
        """
        return self._h5

    @property
    def datasets(self):
        """
        Datasets available

        Returns
        -------
        list
        """
        return self.h5.datasets

    @property
    def dsets(self):
        """
        Datasets available

        Returns
        -------
        list
        """
        return self.datasets

    @property
    def resource_datasets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        return self.h5.resource_datasets

    @property
    def res_dsets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        return self.resource_datasets

    @property
    def shape(self):
        """
        Resource shape (timesteps, sites)
        shape = (len(time_index), len(meta))

        Returns
        -------
        shape : tuple
        """
        return self.h5.shape

    @property
    def meta(self):
        """
        Resource meta data DataFrame

        Returns
        -------
        meta : pandas.DataFrame
        """

        return self.h5.h5.meta

    @property
    def time_index(self):
        """
        Resource DatetimeIndex

        Returns
        -------
        time_index : pandas.DatetimeIndex
        """
        return self.h5.time_index

    @property
    def lat_lon(self):
        """
        Extract (latitude, longitude) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        return self.h5.h5.lat_lon

    @property
    def coordinates(self):
        """
        Coordinates: (lat, lon) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        return self.lat_lon

    @property
    def global_attrs(self):
        """
        Global (file) attributes

        Returns
        -------
        global_attrs : dict
        """
        return self.get_attrs()

    @property
    def attrs(self):
        """
        Dictionary of all dataset attributes

        Returns
        -------
        attrs : dict
        """
        return self.h5.h5.attrs

    @property
    def shapes(self):
        """
        Dictionary of all dataset shapes

        Returns
        -------
        shapes : dict
        """
        return self.h5.h5.shapes

    @property
    def dtypes(self):
        """
        Dictionary of all dataset dtypes

        Returns
        -------
        dtypes : dict
        """
        return self.h5.h5.dtypes

    @property
    def chunks(self):
        """
        Dictionary of all dataset chunk sizes

        Returns
        -------
        chunks : dict
        """
        return self.h5.h5.chunks

    @property
    def scale_factors(self):
        """
        Dictionary of all dataset scale factors

        Returns
        -------
        scale_factors : dict
        """
        return self.h5.h5.scale_factors

    @property
    def units(self):
        """
        Dictionary of all dataset units

        Returns
        -------
        units : dict
        """
        return self.h5.h5.units

    def get_attrs(self, dset=None):
        """
        Get h5 attributes either from file or dataset

        Parameters
        ----------
        dset : str
            Dataset to get attributes for, if None get file (global) attributes

        Returns
        -------
        attrs : dict
            Dataset or file attributes
        """
        return self.h5.h5.get_attrs(dset=dset)

    def get_dset_properties(self, dset):
        """
        Get dataset properties (shape, dtype, chunks)

        Parameters
        ----------
        dset : str
            Dataset to get scale factor for

        Returns
        -------
        shape : tuple
            Dataset array shape
        dtype : str
            Dataset array dtype
        chunks : tuple
            Dataset chunk size
        """
        return self.h5.h5.get_dset_properties(dset)

    def get_scale_factor(self, dset):
        """
        Get dataset scale factor

        Parameters
        ----------
        dset : str
            Dataset to get scale factor for

        Returns
        -------
        float
            Dataset scale factor, used to unscale int values to floats
        """
        return self.h5.h5.get_scale_factor(dset)

    def get_units(self, dset):
        """
        Get dataset units

        Parameters
        ----------
        dset : str
            Dataset to get units for

        Returns
        -------
        str
            Dataset units, None if not defined
        """
        return self.h5.h5.get_units(dset)

    def get_meta_arr(self, rec_name, rows=slice(None)):
        """Get a meta array by name (faster than DataFrame extraction).

        Parameters
        ----------
        rec_name : str
            Named record from the meta data to retrieve.
        rows : slice
            Rows of the record to extract.

        Returns
        -------
        meta_arr : np.ndarray
            Extracted array from the meta data record name.
        """
        meta_arr = self.h5.h5.get_meta_arr(rec_name, rows=rows)

        return meta_arr

    def close(self):
        """
        Close h5 instance
        """
        self._h5.close()


class MultiTimeSolarResource:
    """
    Class to handle solar resource data stored temporaly accross multiple .h5
    files
    """

    def __init__(self, h5_path, unscale=True, str_decode=True, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=SolarResource)


class MultiTimeNSRDB(MultiTimeResource):
    """
    Class to handle NSRDB data stored temporaly accross multiple .h5
    files
    """

    PREFIX = 'nsrdb'

    def __init__(self, h5_path, unscale=True, str_decode=True, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=NSRDB)


class MultiTimeWindResource(MultiTimeResource):
    """
    Class to handle wind resource data stored temporaly accross multiple .h5
    files
    """

    PREFIX = 'wtk'

    def __init__(self, h5_path, unscale=True, str_decode=True, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=WindResource)


class MultiTimeWaveResource(MultiTimeResource):
    """
    Class to handle wave resource data stored temporaly accross multiple .h5
    files
    """

    def __init__(self, h5_path, unscale=True, str_decode=True, hsds=False,
                 hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_path : str | list
            Unix shell style pattern path with * wildcards to multi-file
            resource file sets. Files must have the same coordinates
            but can have different datasets or time indexes. Can also be
            an explicit list of multi time files, which themselves can
            contain * wildcards.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_path, unscale=unscale, hsds=hsds,
                         hsds_kwargs=hsds_kwargs, str_decode=str_decode,
                         res_cls=WaveResource)



# -*- coding: utf-8 -*-
"""
Classes to handle resource data at multiple spatiotemporal resolutions
"""
import pandas as pd
import os
import copy
import logging
from inspect import signature
from scipy.spatial import KDTree

from rex.resource import Resource
from rex.utilities.parse_keys import parse_keys
from rex.utilities.exceptions import ResourceRuntimeError

logger = logging.getLogger(__name__)


class MultiResolutionResource:
    """Multi-resolution resource handler. Uses two resource handlers for files
    at two different spatiotemporal resolutions, and then interpolates the
    lower resolution data to the higher resolution data on the fly.
    """

    INTERPOLABLE_DSETS = ["temperature", "pressure", "windspeed",
                          "winddirection"]
    VARIABLE_NAME = "height"
    VARIABLE_UNIT = "m"

    HR_ATTRS = ('meta', 'time_index', 'coordinates', 'lat_lon', 'data_version',
                'global_attrs', 'get_meta_arr', 'shape')
    """Attributes that are always taken only from the high-res data handler"""

    def __init__(self, h5_hr, h5_lr, handler_class=Resource,
                 handle_kwargs=None, nn_map=None, nn_d=None):
        """
        Parameters
        ----------
        h5_hr : str
            Filepath to high-resolution h5 resource file.
        h5_lr : str
            Filepath to low-resolution h5 resource file.
        handler_class : str
            rex Resource handler class (not initialized) to open both the high
            and low resolution h5 files (both files must be of the same
            resource handler class).
        handle_kwargs : dict, optional
            Dictionary of optional keyword arguments to initialize the
            handler_class for the h5_hr and h5_lr
        nn_map : np.ndarray
            Optional 1D array of nearest neighbor mappings. This will be
            created if not provided. This is created by making a kdtree of the
            lr_res coords and then querying with the hr_res coords. As an
            example, nn_map[10] will return the lr_res index corresponding to
            gid 10 from the hr_res data
        nn_d : np.ndarray
            Optional 1D array of nearest neighbor distances. This will be
            created if not provided. This is created by making a kdtree of the
            lr_res coords and then querying with the hr_res coords. As an
            example, nn_map[10] will return the distance between hr_res gid=10
            and the corresponding lr_res site
        """

        if handle_kwargs is None:
            handle_kwargs = {}

        self._h5_hr = h5_hr
        self._h5_lr = h5_lr
        self._hr_res = handler_class(h5_hr, **handle_kwargs)
        self._lr_res = handler_class(h5_lr, **handle_kwargs)
        self._nn_map = nn_map
        self._nn_d = nn_d

        if self._nn_map is None:
            self._nn_d, self._nn_map = self.make_nn_map(self._hr_res,
                                                        self._lr_res)

        self._interpolation_variable = self._hr_res._parse_interp_var(
            self.datasets)
        self._interpolation_variable.update(self._hr_res._parse_interp_var(
            self.datasets))
        prop_name = "{}s".format(self.VARIABLE_NAME)
        setattr(self, prop_name, self._interpolation_variable)

    @staticmethod
    def make_nn_map(hr_res, lr_res):
        """Make the low-res-to-high-res resource nearest neighbor mapping

        Parameters
        ----------
        hr_res : Resource | MultiFileResource | MultiYearResource
            rex resource handler for the high-resolution data. All retrieval
            gid's are based on this dataset, and the lr_res data is mapped to
            this.
        lr_res : Resource | MultiFileResource | MultiYearResource
            rex resource handler for the low-resolution data. The data from
            this handler is mapped to the hr_res data.

        Returns
        -------
        nn_d : np.ndarray
            1D array of nearest neighbor distances. This is created by making a
            kdtree of the lr_res coords and then querying with the hr_res
            coords. As an example, nn_map[10] will return the distance between
            hr_res gid=10 and the corresponding lr_res site
        nn_map : np.ndarray
            1D array of nearest neighbor mappings. This is created by making a
            kdtree of the lr_res coords and then querying with the hr_res
            coords. As an example, nn_map[10] will return the lr_res index
            corresponding to gid 10 from the hr_res data
        """
        tree = KDTree(lr_res.coordinates)
        nn_d, nn_map = tree.query(hr_res.coordinates)
        return nn_d, nn_map

    def map_ds_slice(self, ds_slice):
        """Map the requested dataset slice from high-res spatial indices to
        low-res spatial indices

        Parameters
        ----------
        ds_slice : tuple
            Tuple where each entry is a slice or list index argument for the
            respective axis, e.g. (slice(None), [0, 2]) retrieves the full
            axis=0 and indices 0 and 2 from axis=1.

        Returns
        -------
        ds_slice : tuple
            Tuple where each entry is a slice or list index argument for the
            respective axis, e.g. (slice(None), [0, 2]) retrieves the full
            axis=0 and indices 0 and 2 from axis=1.
            The returned value is now low-res spatial indices using simple
            nearest neighbor.
        """

        if len(ds_slice) == 1:
            ds_slice = ds_slice + (slice(None), )

        elif len(ds_slice) > 2:
            msg = 'Cannot handle ds_slice > 2D'
            logger.error(msg)
            raise ResourceRuntimeError(msg)

        t_slice, s_slice = ds_slice
        s_slice = self._nn_map[s_slice]
        return (t_slice, s_slice)

    def time_interp(self, arr):
        """Perform temporal interpolation on the low-res data to match the
        high-res data.

        Parameters
        ----------
        arr : np.ndarray
            2D array with shape (time, sites) where time corresponds to the
            low-resolution resource.

        Returns
        -------
        arr : np.ndarray
            2D array with shape (time, sites) where the time axis has been
            linearly interpolated to the high-resolution time index.
        """
        ndim = len(arr.shape)
        arr = pd.DataFrame(arr, index=self._lr_res.time_index)
        arr = arr.reindex(self._hr_res.time_index)
        arr = arr.interpolate('linear').ffill().bfill().values
        if ndim == 1 and len(arr.shape) == 2:
            arr = arr.flatten()
        return arr

    def close(self):
        """Close active file handlers."""
        self._hr_res.close()
        self._lr_res.close()

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self.h5_file)

        return msg

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
        if type is not None:
            raise

    def __len__(self):
        return len(self._hr_res)

    def __getitem__(self, keys):
        ds, ds_slice = parse_keys(keys)
        _, ds_name = os.path.split(ds)
        base_name, _ = self._parse_name(ds_name)

        hr_heights = getattr(self._hr_res, '_interpolation_variable', {})
        lr_heights = getattr(self._lr_res, '_interpolation_variable', {})
        hr_heights = hr_heights.get(base_name, [])
        lr_heights = lr_heights.get(base_name, [])

        if ds_name.startswith('time_index'):
            out = self._hr_res._get_time_index(ds, ds_slice)

        elif ds_name.startswith('meta'):
            out = self._hr_res._get_meta(ds, ds_slice)

        elif ds_name.startswith('coordinates'):
            out = self._hr_res._get_coords(ds, ds_slice)

        elif 'SAM' in ds_name:
            msg = ('SAM dataframe retrieval not implemented for '
                   'MultiResolutionResource, use '
                   'MultiResolutionResource.preload_SAM() instead.')
            logger.error(msg)
            raise NotImplementedError(msg)

        elif ds_name in self._hr_res.dsets or len(hr_heights) > 0:
            out = self._hr_res._get_ds(ds, ds_slice)

        elif ds_name in self._lr_res.dsets or len(lr_heights) > 0:
            ds_slice = self.map_ds_slice(ds_slice)
            out = self._lr_res._get_ds(ds, ds_slice)
            out = self.time_interp(out)

        else:
            msg = ('Could not find data for ds_name = {}. Available dsets in '
                   'hr_file ({}): {}. Available dsets in lr_file ({}): {}.'
                   .format(ds_name, self._h5_hr, self._hr_res.dsets,
                           self._h5_lr, self._lr_res.dsets))
            logger.error(msg)
            raise RuntimeError(msg)

        return out

    def __iter__(self):
        return iter(self.datasets)

    def __contains__(self, dset):
        return dset in self.datasets

    def __getattr__(self, attr):
        if attr in dir(self):
            return getattr(self, attr)
        if attr in self.HR_ATTRS:
            return getattr(self._hr_res, attr)
        else:
            try:
                hr_attr = getattr(self._hr_res, attr)
                lr_attr = getattr(self._lr_res, attr)
                if hasattr(hr_attr, '__call__'):
                    return hr_attr
                elif isinstance(hr_attr, list) and isinstance(lr_attr, list):
                    return list(set(hr_attr + lr_attr))
                elif isinstance(hr_attr, tuple) and isinstance(lr_attr, tuple):
                    return tuple(set(hr_attr + lr_attr))
                elif isinstance(hr_attr, dict) and isinstance(lr_attr, dict):
                    out = copy.deepcopy(lr_attr)
                    out.update(hr_attr)
                    return out
            except Exception as e:
                msg = ('Could not retrieve attribute "{}" from '
                       'MultiResolutionResource handler, the hr and lr '
                       'handler attributes could not be combined: {} {}'
                       .format(attr, hr_attr, lr_attr))
                logger.error(msg)
                raise AttributeError(msg) from e

    @classmethod
    def preload_SAM(cls, h5_hr, h5_lr, sites, *args,
                    handler_class=Resource,
                    nn_map=None, nn_d=None,
                    **kwargs):
        """Pre-load resource data in a SAM resource handler for PySAM / reV run

        Parameters
        ----------
        h5_hr : str
            Filepath to high-resolution h5 resource file.
        h5_lr : str
            Filepath to low-resolution h5 resource file.
        sites : list
            List of sites to be provided to SAM
            (sites is synonymous with gids aka spatial indices)
        *args : list
            Additional arguments required by the resource-specific data handler
            preload_SAM() method (e.g. "hub_heights" is required by
            WindResource handlers and can be provided here).
        handler_class : str
            rex Resource handler class (not initialized) to open both the high
            and low resolution h5 files (both files must be of the same
            resource handler class).
        nn_map : np.ndarray
            Optional 1D array of nearest neighbor mappings. This will be
            created if not provided. This is created by making a kdtree of the
            lr_res coords and then querying with the hr_res coords. As an
            example, nn_map[10] will return the lr_res index corresponding to
            gid 10 from the hr_res data
        nn_d : np.ndarray
            Optional 1D array of nearest neighbor distances. This will be
            created if not provided. This is created by making a kdtree of the
            lr_res coords and then querying with the hr_res coords. As an
            example, nn_map[10] will return the distance between hr_res gid=10
            and the corresponding lr_res site
        **kwargs : dict
            Additional arguments required to either initialize the
            resource-specific data handler or call the resource-specific
            preload_SAM() method (e.g. "hub_heights" is required by
            WindResource.preload_SAM() method and can be provided here in
            addition to optional args like "icing" or "precip_rate").

        Returns
        -------
        SAM_res : SAMResource
            Instance of SAMResource pre-loaded with high-resolution resource
            for sites in project_points
        """

        sig = signature(handler_class)
        handle_kwargs = {k: v for k, v in kwargs.items()
                         if k in sig.parameters}

        cls_kwargs = dict(nn_map=nn_map, nn_d=nn_d,
                          handler_class=handler_class,
                          handle_kwargs=handle_kwargs)

        with cls(h5_hr, h5_lr, **cls_kwargs) as mrr:
            sig = signature(mrr._hr_res._preload_SAM)
            preload_kwargs = {k: v for k, v in kwargs.items()
                              if k in sig.parameters}
            SAM_res = mrr._hr_res._preload_SAM(mrr, sites, *args,
                                               **preload_kwargs)

        return SAM_res



# -*- coding: utf-8 -*-
"""
Module to handle SAM Resource iterator to create site by site resource
DataFrames
"""
from inspect import signature
import numpy as np
import pandas as pd
from warnings import warn
import logging

from rex.utilities.bc_parse_table import parse_bc_table
from rex.utilities.exceptions import (ResourceKeyError, ResourceRuntimeError,
                                      ResourceValueError, SAMInputWarning)
from rex.utilities.parse_keys import parse_keys
from rex.utilities.solar_position import SolarPosition
from rex.utilities.utilities import get_lat_lon_cols

logger = logging.getLogger(__name__)


class SAMResource:
    """
    Resource container for SAM. Resource handlers preload the datasets needed
    by SAM for the sites of interest. SAMResource handles all ETL needed before
    resource data is passed into SAM.

    Examples
    --------
    >>> import os
    >>> from rex import TESTDATADIR, WindResource
    >>> file = os.path.join(TESTDATADIR, 'wtk/ri_100_wtk_2012.h5')

    Here we load a SAM Resource container for a windpower analysis for sites 7
    and 8 at a hub height of 90m:

    >>> sam = WindResource.preload_SAM(file, sites=[7, 8], hub_heights=90)
    >>> sam
    SAMResource with 2 windpower sites

    You can then use the SAMResource object to easily retrieve the data that is
    needed to run the SAM windpower module:

    >>> sam[7]
                               winddirection  pressure  temperature  windspeed
    2012-01-01 00:00:00+00:00            0.0  0.965329        4.270   7.565000
    2012-01-01 01:00:00+00:00            0.0  0.965921        3.870   8.040000
    2012-01-01 02:00:00+00:00            0.0  0.966612        4.070  10.370000
    2012-01-01 03:00:00+00:00            0.0  0.966721        4.060  11.174999
    2012-01-01 04:00:00+00:00            0.0  0.967224        3.515   8.570000
    ...                                  ...       ...          ...        ...
    2012-12-31 19:00:00+00:00            0.0  0.967826       -1.965   6.515000
    2012-12-31 20:00:00+00:00            0.0  0.967036       -2.095   6.750000
    2012-12-31 21:00:00+00:00            0.0  0.966740       -2.495   9.215000
    2012-12-31 22:00:00+00:00            0.0  0.966158       -2.735  10.680000
    2012-12-31 23:00:00+00:00            0.0  0.965852       -2.460  10.805000

    [8784 rows x 4 columns]

    >>> sam['meta', 7]
    latitude         41.975849
    longitude       -71.762329
    country      United States
    state                   RI
    county          Providence
    timezone                -5
    elevation              208
    offshore                 0
    Name: 7, dtype: object

    >>> sam['meta', 8]
    latitude         41.993584
    longitude       -71.754852
    country      United States
    state                   RI
    county          Providence
    timezone                -5
    elevation              180
    offshore                 0
    Name: 8, dtype: object

    >>> sam['meta', 9]
    KeyError: 9
    """

    # Resource variables to load for each SAM technology
    RES_VARS = {'pv': ('dni', 'dhi', 'ghi', 'wind_speed', 'air_temperature'),
                'pvwattsv5': ('dni', 'dhi', 'ghi', 'wind_speed',
                              'air_temperature'),
                'pvwattsv7': ('dni', 'dhi', 'ghi', 'wind_speed',
                              'air_temperature'),
                'pvwattsv8': ('dni', 'dhi', 'ghi', 'wind_speed',
                              'air_temperature'),
                'pvsamv1': ('dni', 'dhi', 'ghi', 'wind_speed',
                            'air_temperature'),
                'csp': ('dni', 'dhi', 'wind_speed', 'air_temperature',
                        'dew_point', 'surface_pressure'),
                'tcsmoltensalt': ('dni', 'dhi', 'wind_speed',
                                  'air_temperature', 'dew_point',
                                  'surface_pressure'),
                'solarwaterheat': ('dni', 'dhi', 'wind_speed',
                                   'air_temperature', 'dew_point',
                                   'surface_pressure'),
                'troughphysicalheat': ('dni', 'dhi', 'wind_speed',
                                       'air_temperature', 'dew_point',
                                       'surface_pressure'),
                'lineardirectsteam': ('dni', 'dhi', 'wind_speed',
                                      'air_temperature', 'dew_point',
                                      'surface_pressure'),
                'wind': ('pressure', 'temperature', 'winddirection',
                         'windspeed'),
                'windpower': ('pressure', 'temperature', 'winddirection',
                              'windspeed'),
                'wave': ('significant_wave_height', 'energy_period'),
                'geothermal': ('temperature', 'potential_MW')}

    # valid data ranges for PV solar resource:
    PV_DATA_RANGES = {'dni': (0.0, 1360.0),
                      'dhi': (0.0, 1360.0),
                      'ghi': (0.0, 1360.0),
                      'wind_speed': (0, 120),
                      'air_temperature': (-200, 100)}

    # valid data ranges for CSP solar resource:
    CSP_DATA_RANGES = {'dni': (0.0, 1360.0),
                       'dhi': (0.0, 1360.0),
                       'ghi': (0.0, 1360.0),
                       'wind_speed': (0, 120),
                       'air_temperature': (-200, 100),
                       'dew_point': (-200, 100),
                       'surface_pressure': (300, 1100)}

    # valid data ranges for wind resource in SAM based on the cpp file:
    # https://github.com/NREL/ssc/blob/develop/shared/lib_windfile.cpp
    WIND_DATA_RANGES = {'windspeed': (0, 120),
                        'winddirection': (0, 360),
                        'pressure': (0.5, 1.099),
                        'temperature': (-200, 100),
                        'rh': (0.1, 99.9)}

    # prevent negative wave data; some negative periods are observed on the
    # west coast along the shore. These are small wave areas and should be fine
    # with setting period to zero. Current limits of pysam (9/2021) cause
    # errors when wave heights or energy periods are greater than the power
    # matrix bin maximums (20.5 and 9.75 respectively)
    WAVE_DATA_RANGES = {'significant_wave_height': (0, 20.5),
                        'energy_period': (0, 9.75)}

    # valid data ranges for trough physical process heat
    TPPH_DATA_RANGES = CSP_DATA_RANGES

    # valid data ranges for linear Fresnel
    LF_DATA_RANGES = CSP_DATA_RANGES

    # valid data ranges for solar water heater
    SWH_DATA_RANGES = CSP_DATA_RANGES

    # valid data ranges for solar water heater
    GEOTHERMAL_DATA_RANGES = {'temperature': (-200, 1000),
                              'potential_MW': (0, 1_000_000)}

    # Data range mapping by SAM tech string
    DATA_RANGES = {'windpower': WIND_DATA_RANGES,
                   'wind': WIND_DATA_RANGES,
                   'pv': PV_DATA_RANGES,
                   'pvwattsv5': PV_DATA_RANGES,
                   'pvwattsv7': PV_DATA_RANGES,
                   'pvwattsv8': PV_DATA_RANGES,
                   'pvsamv1': PV_DATA_RANGES,
                   'csp': CSP_DATA_RANGES,
                   'tcsmoltensalt': CSP_DATA_RANGES,
                   'troughphysicalheat': TPPH_DATA_RANGES,
                   'lineardirectsteam': LF_DATA_RANGES,
                   'solarwaterheat': SWH_DATA_RANGES,
                   'wave': WAVE_DATA_RANGES,
                   'geothermal': GEOTHERMAL_DATA_RANGES}

    # Dataset aliases for flexiblity between NSRDB and WTK naming conventions
    ALIASES = {'wind_speed': 'windspeed',
               'air_temperature': 'temperature'}

    # Variables without a height component that should never be interpolated
    FLAT_VARS = ('dni', 'dhi', 'ghi', 'sza', 'solar_zenith_angle', 'dew_point',
                 'significant_wave_height', 'energy_period')

    def __init__(self, sites, tech, time_index, hub_heights=None, depths=None,
                 require_wind_dir=False, means=False):
        """
        Parameters
        ----------
        sites : int | list | tuple | slice
            int, list, tuple, or slice indicating sites to send to SAM
            (sites is synonymous with gids aka spatial indices)
        tech : str
            SAM technology string. See class attributes for options.
        time_index : pandas.DatetimeIndex
            Time-series datetime index
        hub_heights : int | float | list, optional
            Hub height(s) to extract wind data at, by default None
        depths : int | float | list, optional
            Depth(s) to extract wind data at, by default None
        require_wind_dir : bool, optional
            Boolean flag indicating that wind direction is required,
            by default False
        means : bool, optional
            Boolean flag to compute mean resource when res_array is set,
            by default False
        """
        self._i = 0
        self._sites = self._parse_sites(sites)
        self._time_index = time_index
        self._shape = (len(time_index), len(self._sites))
        self._n = self._shape[1]
        self._var_list = None
        self._meta = None
        self._runnable = False
        self._res_arrays = {}
        self._h = hub_heights
        self._d = depths
        self._sza = None

        self._mean_arrays = None
        if means:
            self._mean_arrays = {}

        if tech.lower() in self.DATA_RANGES:
            self._tech = tech.lower()
        else:
            msg = ('Selected tech {} is not valid. The following technology '
                   'strings are available: {}'
                   .format(tech, list(self.DATA_RANGES.keys())))
            logger.error(msg)
            raise ResourceValueError(msg)

        if self._tech == 'windpower':
            # hub height specified, get WTK wind data.
            if isinstance(self._h, (list, np.ndarray)):
                if len(self._h) != self._n:
                    msg = 'Must have a unique height for each site'
                    logger.error(msg)
                    raise ResourceValueError(msg)

            if not require_wind_dir:
                self._res_arrays['winddirection'] = np.zeros(self._shape,
                                                             dtype='float32')

    def __repr__(self):
        msg = "{} with {} {} sites".format(self.__class__.__name__,
                                           self._n, self._tech)
        return msg

    def __len__(self):
        return self._n

    def __getitem__(self, keys):
        var, var_slice = parse_keys(keys)

        if var == 'time_index':
            out = self.time_index
            out = out[var_slice[0]]
        elif var == 'meta':
            out = self.meta
            out = out.loc[var_slice[0]]
        elif isinstance(var, str):
            if var.startswith('mean_'):
                var = var.replace('mean_', '')
                out = self._get_var_mean(var, *var_slice)
            else:
                out = self._get_var_ts(var, *var_slice)
        elif isinstance(var, int):
            site = var
            out, _ = self._get_res_df(site)
        else:
            msg = 'Cannot interpret {}'.format(var)
            logger.error(msg)
            raise ResourceKeyError(msg)

        return out

    def __setitem__(self, keys, arr):
        var, var_slice = parse_keys(keys)

        if var == 'meta':
            self.meta = arr
        else:
            self._set_var_array(var, arr, *var_slice)

    def __iter__(self):
        return self

    def __next__(self):
        if self._i < self._n:
            site = self.sites[self._i]
            res_df, site_meta = self._get_res_df(site)
            self._i += 1
            return res_df, site_meta
        else:
            raise StopIteration

    @property
    def sites(self):
        """
        Sites being pre-loaded for SAM

        Returns
        -------
        sites : list
            List of sites to be provided to SAM
            (sites is synonymous with gids aka spatial indices)
        """
        sites = self._sites

        return list(sites)

    @property
    def sites_slice(self):
        """Get the sites in slice format if possible

        Returns
        -------
        sites : list | slice
            Sites slice belonging to this instance of ProjectPoints.
            The type is slice if possible. Will be a list only if sites are
            non-sequential.
            (sites is synonymous with gids aka spatial indices)
        """
        # try_slice is what the sites list would be if it is sequential
        if len(self.sites) > 1:
            try_step = self.sites[1] - self.sites[0]
        else:
            try_step = 1

        if try_step <= 0:
            try_step = 1

        try_slice = slice(np.min(self.sites), np.max(self.sites) + 1,
                          try_step)
        try_list = list(range(*try_slice.indices(try_slice.stop)))

        if self.sites == try_list:
            # try_slice is equivelant to the site list
            sites = try_slice
        else:
            # cannot be converted to a sequential slice, return list
            sites = self.sites

        return sites

    @property
    def shape(self):
        """
        Shape of variable arrays

        Returns
        -------
        self._shape : tuple
            Shape (time_index, sites) of variable arrays
        """
        return self._shape

    @property
    def var_list(self):
        """
        Return variable list associated with SAMResource type

        Returns
        -------
        _var_list : list
            List of resource variables associated with resource type
            ('solar' or 'wind')
        """

        if self._var_list is None:
            if self._tech in self.RES_VARS:
                self._var_list = list(self.RES_VARS[self._tech])
            else:
                msg = ("SAM technology string {} is invalid! The following "
                       "technology strings are available: {}"
                       .format(self._tech, list(self.RES_VARS.keys())))
                logging.error(msg)
                raise ResourceValueError(msg)

        return self._var_list

    @property
    def time_index(self):
        """
        Return time_index

        Returns
        -------
        self._time_index : pandas.DatetimeIndex
            Time-series datetime index
        """
        return self._time_index

    @property
    def meta(self):
        """
        Return sites meta

        Returns
        -------
        self._meta : pandas.DataFrame
            DataFrame of sites meta data
        """
        return self._meta

    @meta.setter
    def meta(self, meta):
        """
        Set sites meta

        Parameters
        ----------
        meta : array | pandas.DataFrame
            Sites meta as records array or DataFrame
        """
        if len(meta) != self._n:
            msg = 'Meta does not contain {} sites'.format(self._n)
            logger.error(msg)
            raise ResourceValueError(msg)

        if not isinstance(meta, pd.DataFrame):
            meta = pd.DataFrame(meta, index=self.sites)
        else:
            if not np.array_equal(meta.index, self.sites):
                msg = 'Meta does not match sites!'
                logger.error(msg)
                raise ResourceValueError(msg)

        self._meta = meta

    @property
    def h(self):
        """
        Get heights for wind sites

        Returns
        -------
        self._h : int | float | list
            Hub height or height(s) for wind resource, None for other resource
        """
        return self._h

    @property
    def d(self):
        """
        Get depths for geothermal sites

        Returns
        -------
        self._d : int | float | list
            Depth(s) for geothermal resource, None for other resource
        """
        return self._d

    @property
    def lat_lon(self):
        """
        site latitudes and longitudes

        Returns
        -------
        ndarray
        """
        lat_lon_cols = get_lat_lon_cols(self.meta)
        return self.meta[lat_lon_cols].values

    @property
    def sza(self):
        """
        Solar zenith angle for sites of interest

        Returns
        -------
        ndarray
        """
        if self._sza is None:
            self._sza = \
                np.radians(SolarPosition(self.time_index, self.lat_lon).zenith)

        return self._sza

    @staticmethod
    def _parse_sites(sites):
        """
        Sites to extract resource for and send to SAM

        Parameters
        ----------
        sites : int | list | tuple | slice
            int, list, tuple, or slice indicating sites to send to SAM
            (sites is synonymous with gids aka spatial indices)

        Returns
        -------
        sites : list
            list of sites to send to SAM
            (sites is synonymous with gids aka spatial indices)
        """
        if isinstance(sites, int):
            sites = [sites]
        elif isinstance(sites, slice):
            stop = sites.stop
            if stop is None:
                msg = "sites as a slice must have an explicit stop value!"
                logger.error(msg)
                raise ResourceValueError(msg)

            sites = list(range(*sites.indices(stop)))
        elif not isinstance(sites, (list, tuple)):
            msg = ("sites must a list, tuple or slice, not a {}!"
                   .format(type(slice)))
            logger.error(msg)
            raise ResourceValueError(msg)

        return sites

    @staticmethod
    def check_units(var_name, var_array, tech):
        """
        Check units of variable array and convert to SAM units if needed

        Parameters
        ----------
        var_name : str
            Variable name
        var_array : ndarray
            Variable data
        tech : str
            SAM technology string (windpower, pvwattsv5, solarwaterheat, etc..)

        Returns
        -------
        var_array : ndarray
            Variable data with updated units if needed
        """
        pressure_change = ['csp', 'troughphysicalheat', 'lineardirectsteam',
                           'solarwaterheat']

        if 'pressure' in var_name and tech.lower() == 'windpower':
            # Check if pressure is in Pa, if so convert to atm
            if np.median(var_array) > 1e3:
                # convert pressure from Pa to ATM
                var_array *= 9.86923e-6

        elif 'pressure' in var_name and tech.lower() in pressure_change:
            if np.min(var_array) < 200:
                # convert pressure from 100 to 1000 hPa
                var_array *= 10
            if np.median(var_array) > 70000:
                # convert pressure from Pa to hPa
                var_array /= 100

        elif 'temperature' in var_name and "geothermal" not in tech.lower():
            # Check if tempearture is in K, if so convert to C
            if np.median(var_array) > 200.00:
                var_array -= 273.15

        return var_array

    @staticmethod
    def enforce_arr_range(var, arr, valid_range, sites):
        """Check an array for valid data range, warn, patch, and return.

        Parameters
        ----------
        var : str
            variable name
        arr : np.ndarray
            Array to be checked and patched
        valid_range : np.ndarray | tuple | list
            arr data will be ensured within the min/max values of valid_range
        sites : list
            Resource gid site list for warning printout.
            (sites is synonymous with gids aka spatial indices)

        Returns
        -------
        arr : np.ndarray
            Patched array with valid range.
        """
        min_val = np.min(valid_range)
        max_val = np.max(valid_range)
        check_low = (arr < min_val)
        check_high = (arr > max_val)
        check = (check_low | check_high)
        if check.any():
            warn('Resource dataset "{}" out of viable SAM range ({}, {}) for '
                 'sites {}. Data min/max: {}/{}. Patching data...'
                 .format(var, min_val, max_val,
                         list(np.array(sites)[check.any(axis=0)]),
                         np.min(arr), np.max(arr)),
                 SAMInputWarning)

            arr[check_low] = min_val
            arr[check_high] = max_val

        return arr

    @staticmethod
    def roll_timeseries(time_series, timezone, time_interval):
        """
        Roll timeseries array to given timezone from UTC

        Parameters
        ----------
        time_series : ndarray
            time_series array to roll
        timezone : int
            Time zone as UTC offset
        time_interval : int
            Number of step-steps in an hour, needed to compute time shift

        Returns
        -------
        time_series : ndarray
            Time series in local time
        """
        shift = int(timezone * time_interval)
        time_series = np.roll(time_series, shift, axis=0)

        return time_series

    def check_irradiance_datasets(self, datasets, clearsky=False):
        """
        Check available irradiance datasets

        Parameters
        ----------
        datasets : list
            List of available datasets in resource .h5 file
        clearsky : bool, optional
            Flag to check for clearsky irradiance datasets, by default False
        """
        available = 0
        irradiance_vars = ['dni', 'dhi', 'ghi']
        if clearsky:
            irradiance_vars = ['clearsky_{}'.format(var)
                               for var in irradiance_vars]
        for var in irradiance_vars:
            if var in datasets and var in self.var_list:
                available += 1

        if available < 2:
            msg = ("At least 2 irradiance variables (dni, dhi, or ghi) are "
                   "needed to run SAM!")
            logger.error(msg)
            raise ResourceRuntimeError(msg)

    def compute_irradiance(self, clearsky=False):
        """
        Fillin missing irradiance dataset from available values and SZA

        Parameters
        ----------
        clearsky : bool, optional
            Flag to check for clearsky irradiance datasets, by default False
        """
        irradiance_vars = ['dni', 'dhi', 'ghi']
        if clearsky:
            irradiance_vars = ['clearsky_{}'.format(var)
                               for var in irradiance_vars]

        missing = None
        for var in irradiance_vars:
            if var in self.var_list and var not in self._res_arrays:
                missing = var
                break

        if missing is not None:
            dni_var, dhi_var, ghi_var = irradiance_vars
            logger.info('{} is missing and will be computed from {}'
                        .format(missing, irradiance_vars.remove(missing)))
            if missing == ghi_var:
                ghi = (self._res_arrays[dni_var] * np.cos(self.sza)
                       + self._res_arrays[dhi_var])
                ghi[ghi < 0] = 0
                self[ghi_var] = ghi
            elif missing == dni_var:
                dni = ((self._res_arrays[ghi_var] - self._res_arrays[dhi_var])
                       / np.cos(self.sza))
                dni = np.nan_to_num(dni)
                dni[dni < 0] = 0
                self[dni_var] = dni
            elif missing == dhi_var:
                dhi = (self._res_arrays[ghi_var]
                       - self._res_arrays[dni_var] * np.cos(self.sza))
                dhi[dhi < 0] = 0
                self[dhi_var] = dhi

    def set_clearsky(self):
        """Make the NSRDB var list for solar based on clearsky irradiance."""
        for i, var in enumerate(self.var_list):
            if var in ['dni', 'dhi', 'ghi']:
                self._var_list[i] = 'clearsky_{}'.format(var)

    def append_var_list(self, var):
        """
        Append a new variable to the SAM resource protected var_list.

        Parameters
        ----------
        var : str
            New resource variable to be added to the protected var_list
            property.
        """

        self.var_list.append(var)

    def bias_correct(self, bc_df):
        """Bias correct wind or irradiance data using a table of linear
        correction factors per resource gid.

        Parameters
        ----------
        bc_df : pd.DataFrame
            DataFrame with wind or solar resource bias correction table. This
            must have columns "gid" and "method", where "gid" is the resource
            file indices, and "method" is a function name from the
            ``rex.bias_correction`` module. Only windspeed or GHI+DNI+DHI are
            corrected, depending on the technology. See the
            ``rex.bias_correction`` module for more details on available
            bias correction methods.
        """

        bc_fun, bc_fun_kwargs, bool_bc = parse_bc_table(bc_df, self.sites)

        if not bool_bc.any():
            return

        if 'ghi' in self._res_arrays and 'dni' in self._res_arrays:
            logger.debug('Bias correcting irradiance with function {} '
                         'for sites {}'.format(bc_fun, self.sites))
            ghi = self._res_arrays['ghi']
            dni = self._res_arrays['dni']
            dhi = self._res_arrays['dhi']

            bc_fun_kwargs['ghi'] = ghi[:, bool_bc]
            bc_fun_kwargs['dni'] = dni[:, bool_bc]
            bc_fun_kwargs['dhi'] = dhi[:, bool_bc]

            sig = signature(bc_fun)
            bc_fun_kwargs = {k: v for k, v in bc_fun_kwargs.items()
                             if k in sig.parameters}
            out = bc_fun(**bc_fun_kwargs)

            ghi[:, bool_bc] = out[0][:, bool_bc]
            dni[:, bool_bc] = out[1][:, bool_bc]
            dhi[:, bool_bc] = out[2][:, bool_bc]

            self._res_arrays['ghi'] = ghi
            self._res_arrays['dni'] = dni
            self._res_arrays['dhi'] = dhi

        elif 'windspeed' in self._res_arrays:
            logger.debug('Bias correcting windspeed with function {} '
                         'for sites {}'.format(bc_fun, self.sites))
            ws = self._res_arrays['windspeed']
            bc_fun_kwargs['ws'] = ws[:, bool_bc]
            sig = signature(bc_fun)
            bc_fun_kwargs = {k: v for k, v in bc_fun_kwargs.items()
                             if k in sig.parameters}
            ws[:, bool_bc] = bc_fun(**bc_fun_kwargs)
            self._res_arrays['windspeed'] = ws

        if self._mean_arrays is not None:
            # pylint: disable=consider-iterating-dictionary
            for var in self._mean_arrays.keys():
                self._mean_arrays[var] = self._res_arrays[var].mean(axis=0)

    def _check_physical_ranges(self, var, arr, var_slice):
        """Check physical range of array and enforce usable SAM data.

        Parameters
        ----------
        var : str
            variable name
        arr : np.ndarray
            Array to be checked and patched
        var_slice : tuple of int | list | slice
            Slice of variable array to extract

        Returns
        -------
        arr : np.ndarray
            Patched array with valid range.
        """

        # Get site list corresponding to the var_slice. Only reduce the sites
        # list if the var_slice has a second entry (column slice of sites)
        arr_sites = self.sites
        if not isinstance(var_slice, slice):
            if (len(var_slice) > 1
                    and not isinstance(var_slice[1], slice)):
                arr_sites = list(np.array(self.sites)[np.array(var_slice[1])])

        if var in self.DATA_RANGES[self._tech]:
            valid_range = self.DATA_RANGES[self._tech][var]
            arr = self.enforce_arr_range(var, arr, valid_range, arr_sites)

        return arr

    def runnable(self):
        """
        Check to see if SAMResource iterator is runnable:
        - Meta must be loaded
        - Variables in var_list must be loaded

        Returns
        ------
        bool
            Returns True if runnable check passes
        """
        if self._meta is None:
            msg = 'meta has not been set!'
            logger.error(msg)
            raise ResourceRuntimeError(msg)
        else:
            for var in self.var_list:
                if var not in self._res_arrays:
                    msg = '{} has not been set!'.format(var)
                    logger.error(msg)
                    raise ResourceRuntimeError(msg)

        return True

    def _set_var_array(self, var, arr, *var_slice):
        """
        Set variable array (units and physical ranges are checked while set).

        Parameters
        ----------
        var : str
            Resource variable name
        arr : ndarray
            Time series data of given variable for sites
        var_slice : tuple of int | list | slice
            Slice of variable array that corresponds to arr
        """
        if var in self.var_list:
            var_arr = self._res_arrays.get(var, np.zeros(self._shape,
                                                         dtype='float32'))
            if var_arr[var_slice].shape == arr.shape:
                arr = self.check_units(var, arr, self._tech)
                arr = self._check_physical_ranges(var, arr, var_slice)
                var_arr[var_slice] = arr
                self._res_arrays[var] = var_arr
                if self._mean_arrays is not None:
                    self._mean_arrays[var] = var_arr.mean(axis=0)
            else:
                msg = ('{} has shape {}, '
                       'needs proper shape: {}'.format(var,
                                                       arr.shape, self._shape))
                logger.error(msg)
                raise ResourceValueError(msg)
        else:
            msg = '{} not in {}'.format(var, self.var_list)
            logger.error(msg)
            raise ResourceKeyError(msg)

    def _get_var_mean(self, var, *var_slice):
        """
        Get variable means

        Parameters
        ----------
        var : str
            Resource variable name
        var_slice : int | list | slice
            Slice of variable array to extract

        Returns
        -------
        means : ndarray
            Vector of variable means
        """
        if self._mean_arrays is None:
            msg = ("Variable means were not computed, ensure ws_mean for "
                   "windpower, or dni_mean/ghi_mean for pvwatts is in "
                   "'output_request'")
            logger.error(msg)
            raise ResourceRuntimeError(msg)

        if var in self.var_list:
            try:
                var_array = self._mean_arrays[var]
            except KeyError as ex:
                msg = '{} has yet to be set!'.format(var)
                logger.error(msg)
                raise ResourceKeyError(msg) from ex

            means = var_array[var_slice]
        else:
            msg = '{} not in {}'.format(var, self.var_list)
            logger.error(msg)
            raise ResourceKeyError(msg)

        return means

    def _get_var_ts(self, var, *var_slice):
        """
        Get variable time-series

        Parameters
        ----------
        var : str
            Resource variable name
        var_slice : tuple of int | list | slice
            Slice of variable array to extract

        Returns
        -------
        ts : pandas.DataFrame
            Time-series for desired sites of variable var
        """
        if var in self.var_list:
            try:
                var_array = self._res_arrays[var]
            except KeyError as ex:
                msg = '{} has yet to be set!'.format(var)
                logger.error(msg)
                raise ResourceKeyError(msg) from ex

            sites = np.array(self.sites)
            if len(var_slice) == 2:
                sites = sites[var_slice[1]]

            ts = pd.DataFrame(var_array[var_slice],
                              index=self.time_index[var_slice[0]],
                              columns=sites)
        else:
            msg = '{} not in {}'.format(var, self.var_list)
            logger.error(msg)
            raise ResourceKeyError(msg)

        return ts

    def _get_res_df(self, site):
        """
        Get resource time-series

        Parameters
        ----------
        site : int
            Site to extract

        Returns
        -------
        res_df : pandas.DataFrame
            Time-series of SAM resource variables for given site
        site_meta : pandas.Series
            Meta data for the input site
        """

        self.runnable()
        try:
            idx = self.sites.index(site)
        except ValueError as ex:
            msg = '{} is not in available sites'.format(site)
            logger.error(msg)
            raise ResourceValueError(msg) from ex

        site_meta = self.meta.loc[site].copy()
        if not isinstance(site_meta, pd.Series):
            site_meta = site_meta.iloc[0]

        if self._h is not None:
            try:
                h = self._h[idx]
            except TypeError:
                h = self._h

            site_meta['height'] = h

        res_df = pd.DataFrame(index=self.time_index)
        res_df.name = site
        for var_name, var_array in self._res_arrays.items():
            res_df[var_name] = var_array[:, idx]

        return res_df, site_meta

    def curtail_windspeed(self, gids, curtailment):
        """
        Apply temporal curtailment mask to windspeed resource at given sites

        Parameters
        ----------
        gids : int | list
            gids for site or list of sites to curtail
        curtailment : ndarray
            Temporal multiplier for curtailment
        """
        shape = (self.shape[0],)
        if isinstance(gids, int):
            site_pos = self.sites.index(gids)
        else:
            shape += (len(gids),)
            site_pos = [self.sites.index(id) for id in gids]

        if curtailment.shape != shape:
            msg = "curtailment must be of shape: {}".format(shape)
            logger.error(msg)
            raise ResourceValueError(msg)

        if 'windspeed' in self._res_arrays:
            self._res_arrays['windspeed'][:, site_pos] *= curtailment
        else:
            msg = 'windspeed has not be loaded!'
            logger.error(msg)
            raise ResourceRuntimeError(msg)

    @staticmethod
    def _check_site_request(rex_res, sites):
        """
        Parameters
        ----------
        rex_res : rex.Resource
            rex Resource handler or similar (NSRDB, WindResource,
            MultiFileResource, etc...)
        sites : list | slice | int
            List of site indices (axis=1)
            (sites is synonymous with gids aka spatial indices)
        """
        if isinstance(sites, slice):
            last = np.arange(sites.stop)[sites][-1]
        elif isinstance(sites, (list, tuple)):
            last = sorted(sites)[-1]

        if last > rex_res.shape[1] - 1:
            msg = ('Cannot retrieve site index {} from rex resource of '
                   'shape {}: {}'
                   .format(last, rex_res.shape, rex_res))
            logger.error(msg)
            raise ResourceKeyError(msg)

    def load_rex_resource(self, rex_res, var_list, time_slice, sites, hh=None,
                          hh_unit='m'):
        """Load data from a rex Resource handler into this SAMResource
        container.

        Parameters
        ----------
        rex_res : rex.Resource
            rex Resource handler or similar (NSRDB, WindResource,
            MultiFileResource, etc...)
        var_list : list
            List of variables to retrieve from rex_res. These names may be
            manipulated with suffixes such as _100m (for a 100m hh input)
        time_slice : slice
            Slicing argument for the resource temporal dimension (axis=0)
        sites : list | slice | int
            List of site indices (axis=1)
            (sites is synonymous with gids aka spatial indices)
        hh : None | int
            Optional single hub height in meters that datasets are to be loaded
            from rex_res at
        hh_unit : str
            Unit suffix for the hub height input.
        """

        self._check_site_request(rex_res, sites)

        for sam_var in var_list:
            alias = self.ALIASES.get(sam_var, None)
            var_hh = "{}_{}{}".format(sam_var, hh, hh_unit)
            alias_hh = "{}_{}{}".format(alias, hh, hh_unit)
            use_hh = sam_var not in self.FLAT_VARS

            res_var = sam_var
            if res_var in rex_res.datasets:
                pass
            elif alias in rex_res.datasets:
                res_var = alias
            elif hh is not None and alias is None and use_hh:
                res_var = var_hh
            elif hh is not None and alias is not None and use_hh:
                res_var = alias_hh

            try:
                arr = rex_res[res_var, time_slice, sites]
                self._set_var_array(sam_var, arr)
            except ResourceKeyError as e:
                msg = ('Could not get SAM resource "{}" with retrieval '
                       'dataset "{}" from rex Resource handler: {}, received '
                       'error: {}'
                       .format(sam_var, res_var, rex_res, e))
                logger.warning(msg)
                warn(msg)



"""
setup.py
"""
import os
from codecs import open
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from subprocess import check_call
import shlex
from warnings import warn

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "rex", "version.py"), encoding="utf-8") as f:
    version = f.read()

version = version.split('=')[-1].strip().strip('"').strip("'")


class PostDevelopCommand(develop):
    """
    Class to run post setup commands
    """

    def run(self):
        """
        Run method that tries to install pre-commit hooks
        """
        try:
            check_call(shlex.split("pre-commit install"))
        except Exception as e:
            warn("Unable to run 'pre-commit install': {}"
                 .format(e))

        develop.run(self)


with open("requirements.txt") as f:
    install_requires = f.readlines()

test_requires = ["pytest>=5.2", "pytest-timeout>=2.3.1", "flaky>=3.8.1"]
dev_requires = ["flake8", "pre-commit", "pylint", "hsds>=0.8.4"]
description = ("National Renewable Energy Laboratory's (NREL's) REsource "
               "eXtraction tool: rex")

setup(
    name="NREL-rex",
    version=version,
    description=description,
    long_description=readme,
    author="Michael Rossol",
    author_email="michael.rossol@nrel.gov",
    url="https://nrel.github.io/rex/",
    packages=find_packages(),
    package_dir={"rex": "rex"},
    entry_points={
        "console_scripts": [
            "rex=rex.resource_extraction.resource_cli:main",
            "NSRDBX=rex.resource_extraction.nsrdb_cli:main",
            "WINDX=rex.resource_extraction.wind_cli:main",
            "WaveX=rex.resource_extraction.wave_cli:main",
            "MultiYearX=rex.resource_extraction.multi_year_resource_cli:main",
            "US-wave=rex.resource_extraction.US_wave_cli:main",
            "rechunk=rex.rechunk_h5.rechunk_cli:main",
            "combine-h5=rex.rechunk_h5.combine_h5_cli:main",
            "temporal-stats=rex.temporal_stats.temporal_stats_cli:main",
            "wind-rose=rex.joint_pd.wind_rose_cli:main"
        ],
    },
    include_package_data=True,
    license="BSD 3-Clause",
    zip_safe=False,
    keywords="rex",
    python_requires='>=3.8',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="tests",
    install_requires=install_requires,
    extras_require={
        "test": test_requires,
        "dev": test_requires + dev_requires,
        "hsds": ["hsds>=0.8.4"],
        "s3": ['fsspec', 's3fs'],
    },
    cmdclass={"develop": PostDevelopCommand},
)
